
/////////////////////////////////////////////////////////////////////
// DESP-C++ (Discrete-Event Simulation Package)
// Version 1.1, February 1998
// Jerome Darmont
// LIMOS, Blaise Pascal University (Clermont-Ferrand II), France
/////////////////////////////////////////////////////////////////////
// eventc.h: Event manager classes definition
// Varies with the simulated system
// VOODB2 set-up (April 1998)
/////////////////////////////////////////////////////////////////////

// Active resources declaration 

class User;
class TransMgr;
class ClustMgr;
class ObjMgr;
class BuffMgr;
class IOSubSys;

/////////////////////////////////////////////////////////////////////
// CLASS EventManager
/////////////////////////////////////////////////////////////////////
// Simulation events management
/////////////////////////////////////////////////////////////////////

// The Event Manager must "know" all (active and passive) resources.

class EventManager {

  public:

    // Methods

    EventManager(Simulation *sim);      // Constructor
    ~EventManager();                    // Destructor
    void ExecuteEvent(int code, Client *client); // Event execution
    void Init();                        // Initialization
    void InitRep();                     // Replication initialization 
    void Stats();                       // Stats computation (end of replication)
    void DisplayStats();                // Statistics display

  private:

    // Attributes

    Simulation *simul;                  // Ptr toward simulation object

    // Passive resources 
    Resource *serverpr;                 // Server processor & main memory
    Resource *clientpr[MNUSERS];        // Clients processor & main memory
    Resource *disk;                     // Server disk controler & disk
    Resource *db;                       // Database (concurrency control)

    // Active resources 
    User *user[MNUSERS];
    TransMgr *transmgr;
    ClustMgr *clustmgr;
    ObjMgr *objmgr;
    BuffMgr *buffmgr;
    IOSubSys *iosubsys;

};

/////////////////////////////////////////////////////////////////////
// CLASS Client
/////////////////////////////////////////////////////////////////////
// Custom simulation entity
/////////////////////////////////////////////////////////////////////

class Client {

  public:
           // Add here (in public) additional attributes
           int cliid;           // Client ID
           short type;          // Client type
                                // 0 = Transaction
                                // 1 = Set-oriented access
                                // 2 = Simple traversal
                                // 3 = Hierarchy traversal
                                // 4 = Stochastic traversal
                                // 5 = Clustering demand
                                // 7 = Direct writing on disk
                                // 9 = Fictitious client
           // Transaction
           int rootoid;         // Root object
           int trans_it;        // Number of iteration in transaction
           int prevoid;         // Previously accessed object
           int curdepth;        // Current depth
           ChList *roots;       // List of "roots"
           ChList *toaccess;    // List of objects to access
           // Hierarchy traversal
           short reftype;       // Reference type
           // Page or I/O request
           ChList *pgnums;      // Page numbers
           int pageindex;       // Page index
           // Clustering demand
           ChList *o1;          // List of object couples (o1,o2)
           ChList *o2;          // to cluster
           ChList *fc;          // with link weight (fc)

           // Base attributes

           Client *next;
           Client *previous;

};

/////////////////////////////////////////////////////////////////////
// CLASS User
/////////////////////////////////////////////////////////////////////
// AR1: User
/////////////////////////////////////////////////////////////////////

class User: public Resource {

  public:

    // Constructor

    User(char n[STRS], int cap, Simulation *sim, short id, Resource *rsp);

    // Events

    void Event10(Client *client);       // Arrival on AR1
    void Event11(Client *client);       // Latency time
    void Event12(Client *client);       // Wait for processor
    void Event13(Client *client);       // Transaction generation

  private:

    // Identifier

    short userid;

    // Resource pointers

    Resource *rsproc;

    // Control subsystem

    Client *GenTrans();                 // R1

};


/////////////////////////////////////////////////////////////////////
// CLASS TransMgr
/////////////////////////////////////////////////////////////////////
// AR2: Transaction Manager
/////////////////////////////////////////////////////////////////////

class TransMgr: public Resource {

  public:

    // Constructor

    TransMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsb, float t);

    // Events

    void Event20(Client *client);       // Arrival on AR2
    void Event21(Client *client);       // Wait for processor (1)
    void Event22(Client *client);       // Extract object / reserve DB
    void Event23(Client *client);       // Wait for processor (2)
    void Event24(Client *client);       // Acquire locks
    void Event25(Client *client);       // Object request
    void Event26(Client *client);       // Wait for processor (3)
    void Event27(Client *client);       // Transaction treatment & lock release
    void Event28(Client *client);       // Network transfer

  private:

    // Network transfer time

    float tnet;

    // Resource pointers

    Resource *rsproc[MNUSERS], *rsbd;

    // Control subsystem

    void ExtObj(Client *client);        // R2
    void NextSet(Client *client);       // R2.1: Set-oriented access
    void NextSimp(Client *client);      // R2.2: Simple traversal
    void NextHier(Client *client);      // R2.3: Hierarchy traversal
    void NextStoc(Client *client);      // R2.4: Stochastic traversal
    void TrTrans();                     // R6

};

/////////////////////////////////////////////////////////////////////
// CLASS ClustMgr
/////////////////////////////////////////////////////////////////////
// AR3: Clustering Manager
/////////////////////////////////////////////////////////////////////

class ClustMgr: public Resource {

  public:

    // Constructor

    ClustMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsb);
    void SetBuffPtr(BuffMgr *ptr);      // Initialize pointer toward BuffMgr
    void SetObjPtr(ObjMgr *ptr);        // Initialize pointer toward ObjMgr

    // Events

    void Event30(Client *client);       // Reservation AR3 & Treatment selection or clustering

    void Event31(Client *client);       // Tt: Wait for processor
    void Event32(Client *client);       // Tt: R7
    void Event33(Client *client);       // Tt: End of treatment

    void Event34(Client *client);       // Cl : Lock DB
    void Event35(Client *client);       // Cl : Destroy fictitious clients 
    void Event36(Client *client);       // Cl : Wait for processor
    void Event37(Client *client);       // Cl : R8
    void Event38(Client *client);       // Cl : End of clustering

  private:

    // Resource pointers

    Resource *rsproc[MNUSERS], *rsbd;
    ObjMgr *objmgr;
    BuffMgr *buffmgr;

    // Control subsystem

    float ClustTt(Client *client);      // R7, return time needed
    float Cluster(Client *client);      // R8, return time needed
    float DSTC(Client *client);         // DSTC clustering, return time needed
    ChList *Schedule(ChList *ur[3]);    // DSTC: Ordering of clustering units into clusters
    float ObjRequest(Client *client);   // Object request, return time needed
    float AccDisk(int pagenum);         // Direct disk access, return time needed

    // Private attributes

    int lastpage;                       // Last page accessed when clustering

    ChList *Clust(ChList *ur[3]) {
      ChList *cluster=new ChList;
      int i,j,k,v;
      short trouve;
      for (k=0;k<2;k++)
        for (i=1;i<=ur[k]->NbElem();i++) {
          v=ur[k]->Scan(i);
          trouve=0;
          j=1;
          while ((!trouve)&&(j<=cluster->NbElem()))
            if (cluster->Scan(j)==v) trouve=1;
            else j++;
          if (!trouve) cluster->Insert(1,v);
        }
      return cluster;
    }
};

// Class Node
// Node in DSTC clustering unit ordering algorithm

class Node {
  public:
    int num;                      // Node number
    ChList *list;                 // List of node objects 
    Node *next;                   // Ptr toward next node
};

/////////////////////////////////////////////////////////////////////
// CLASS ObjMgr
/////////////////////////////////////////////////////////////////////
// AR4: Object Manager
/////////////////////////////////////////////////////////////////////

class ObjMgr: public Resource {

  public:

    // Constructor

    ObjMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, float t);

    // Information

    float Tnet();                       // Return tnet

    // Events

    void Event40(Client *client);       // Reservation AR4
    void Event41(Client *client);       // Wait for processor
    void Event42(Client *client);       // R3 + prefetching
    void Event43(Client *client);       // Object in memory

    // Control subsystem

    void ExtPg(Client *client);         // R3
    void Prefetch(Client *client);      // Prefetching routine

  private:

    // Resource pointers

    Resource *rsproc[MNUSERS];

    // Unit of network time

    float tnet;

};

/////////////////////////////////////////////////////////////////////
// CLASS BuffMgr
/////////////////////////////////////////////////////////////////////
// AR5: Buffering Manager
/////////////////////////////////////////////////////////////////////

// Class BuffSlot: "cell" in buffer

class BuffSlot {

public: int pgnum;              // Page number
        short updated;          // Update flag 
        int nbacc;              // Number of accesses (LFU, GCLOCK)
        float date[MKLRU];      // Last usage date (LRU)
        short clockflag;        // Usage flag (CLOCK)
};

class BuffMgr: public Resource {

  public:

    // Constructor

    BuffMgr(char n[STRS], int cap, Simulation *sim, Resource *rsp, float t);

    // Information

    float Tnet();                       // Return tnet
    short Updated(int num);             // Return 1 if page #num was "updated", 0 otherwise
    int Page(int num);                  // Return number of page in position #num

    // Buffer modifications

    void InitBuffer();                  // Buffer initialization (empty)
    void SetUpdated(int num);           // Mark page #num to "updated"
    void SetBuffer(int num,int pgnum,float time); // Insert page #pgnum in position #num at time time

    // Events

    void Event50(Client *client);       // Reservation AR5
    void Event51(Client *client);       // Wait for processor (1)
    void Event52(Client *client);       // R4 & I/O request if necessary
    void Event53(Client *client);       // Wait for processor (2)
    void Event54(Client *client);       // Page replacement 
    void Event55(Client *client);       // Network transfer

    // Control subsystem public part

    void AccPage(Client *client);       // R4
    int Replace();                      // Buffer page replacement 

  private:

    // Resource pointers

    Resource *rsproc;

    // Unit of network time

    float tnet;

    // Buffer definition

    BuffSlot buffer[MBUFFSIZE];         // Buffer
    int buffpos;                        // Buffer filling position
    int clock;                          // Position saving / algorithms
                                        // LFU, LRU, CLOCK, GCLOCK

    // Control subsystem

    int RANDOMrep();                    // Replacement: RANDOM
    int FIFOrep();                      // Replacement: FIFO
    int LFUrep();                       // Replacement: LFU
    int LRUrep();                       // Replacement: LRU-K
    int CLOCKrep();                     // Replacement: CLOCK
    int GCLOCKrep();                    // Replacement: GCLOCK
    int Otherep();                      // Replacement: other (user-defined)

};

/////////////////////////////////////////////////////////////////////
// CLASS IOSubSys
/////////////////////////////////////////////////////////////////////
// AR6: I/O Subsystem
/////////////////////////////////////////////////////////////////////

class IOSubSys: public Resource {

  public:

    // Constructor

    IOSubSys(char n[STRS], int cap, Simulation *sim, Resource *rsp, Resource *rsd);

    // Events

    void Event60(Client *client);       // Reservation AR6
    void Event61(Client *client);       // Wait for processor
    void Event62(Client *client);       // Wait for disk
    void Event63(Client *client);       // Access disk (R5)

  private:

    // Number of last page accessed

    int lastpage;

    // Resource pointers

    Resource *rsproc, *rsdisk;

    // Control subsystem

    void AccDisk(Client *client);       // R5

};


